/*
 * Copyright 2017, Matt Brown, IBM Corp.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * vpermxor8.c
 *
 * Based on H. Peter Anvin's paper - The mathematics of RAID-6
 *
 * 8-way unrolled portable integer math RAID-6 instruction set
 * This file is postprocessed using unroll.awk
 *
 * vpermxor8.c makes use of the vpermxor instruction to optimise the RAID6 Q
 * syndrome calculations.
 * This can be run on systems which have both Altivec and vpermxor instruction.
 *
 * This instruction was introduced in POWER8 - ISA v2.07.
 */

#include <linux/raid/pq.h>
#ifdef CONFIG_ALTIVEC

#include <altivec.h>
#ifdef __KERNEL__
#include <asm/cputable.h>
#include <asm/ppc-opcode.h>
#include <asm/switch_to.h>
#endif

typedef vector unsigned char unative_t;
#define NSIZE sizeof(unative_t)

static const vector unsigned char gf_low = {0x1e, 0x1c, 0x1a, 0x18, 0x16, 0x14,
					    0x12, 0x10, 0x0e, 0x0c, 0x0a, 0x08,
					    0x06, 0x04, 0x02,0x00};
static const vector unsigned char gf_high = {0xfd, 0xdd, 0xbd, 0x9d, 0x7d, 0x5d,
					     0x3d, 0x1d, 0xe0, 0xc0, 0xa0, 0x80,
					     0x60, 0x40, 0x20, 0x00};

static void noinline raid6_vpermxor8_gen_syndrome_real(int disks, size_t bytes,
							void **ptrs)
{
	u8 **dptr = (u8 **)ptrs;
	u8 *p, *q;
	int d, z, z0;
	unative_t wp0, wq0, wd0;
	unative_t wp1, wq1, wd1;
	unative_t wp2, wq2, wd2;
	unative_t wp3, wq3, wd3;
	unative_t wp4, wq4, wd4;
	unative_t wp5, wq5, wd5;
	unative_t wp6, wq6, wd6;
	unative_t wp7, wq7, wd7;

	z0 = disks - 3;		/* Highest data disk */
	p = dptr[z0+1];		/* XOR parity */
	q = dptr[z0+2];		/* RS syndrome */

	for (d = 0; d < bytes; d += NSIZE*8) {
		wp0 = wq0 = *(unative_t *)&dptr[z0][d+0*NSIZE];
		wp1 = wq1 = *(unative_t *)&dptr[z0][d+1*NSIZE];
		wp2 = wq2 = *(unative_t *)&dptr[z0][d+2*NSIZE];
		wp3 = wq3 = *(unative_t *)&dptr[z0][d+3*NSIZE];
		wp4 = wq4 = *(unative_t *)&dptr[z0][d+4*NSIZE];
		wp5 = wq5 = *(unative_t *)&dptr[z0][d+5*NSIZE];
		wp6 = wq6 = *(unative_t *)&dptr[z0][d+6*NSIZE];
		wp7 = wq7 = *(unative_t *)&dptr[z0][d+7*NSIZE];

		for (z = z0-1; z>=0; z--) {
			wd0 = *(unative_t *)&dptr[z][d+0*NSIZE];
			wd1 = *(unative_t *)&dptr[z][d+1*NSIZE];
			wd2 = *(unative_t *)&dptr[z][d+2*NSIZE];
			wd3 = *(unative_t *)&dptr[z][d+3*NSIZE];
			wd4 = *(unative_t *)&dptr[z][d+4*NSIZE];
			wd5 = *(unative_t *)&dptr[z][d+5*NSIZE];
			wd6 = *(unative_t *)&dptr[z][d+6*NSIZE];
			wd7 = *(unative_t *)&dptr[z][d+7*NSIZE];
			/* P syndrome */
			wp0 = vec_xor(wp0, wd0);
			wp1 = vec_xor(wp1, wd1);
			wp2 = vec_xor(wp2, wd2);
			wp3 = vec_xor(wp3, wd3);
			wp4 = vec_xor(wp4, wd4);
			wp5 = vec_xor(wp5, wd5);
			wp6 = vec_xor(wp6, wd6);
			wp7 = vec_xor(wp7, wd7);

			/* Q syndrome */
			asm(VPERMXOR(%0,%1,%2,%3):"=v"(wq0):"v"(gf_high), "v"(gf_low), "v"(wq0));
			asm(VPERMXOR(%0,%1,%2,%3):"=v"(wq1):"v"(gf_high), "v"(gf_low), "v"(wq1));
			asm(VPERMXOR(%0,%1,%2,%3):"=v"(wq2):"v"(gf_high), "v"(gf_low), "v"(wq2));
			asm(VPERMXOR(%0,%1,%2,%3):"=v"(wq3):"v"(gf_high), "v"(gf_low), "v"(wq3));
			asm(VPERMXOR(%0,%1,%2,%3):"=v"(wq4):"v"(gf_high), "v"(gf_low), "v"(wq4));
			asm(VPERMXOR(%0,%1,%2,%3):"=v"(wq5):"v"(gf_high), "v"(gf_low), "v"(wq5));
			asm(VPERMXOR(%0,%1,%2,%3):"=v"(wq6):"v"(gf_high), "v"(gf_low), "v"(wq6));
			asm(VPERMXOR(%0,%1,%2,%3):"=v"(wq7):"v"(gf_high), "v"(gf_low), "v"(wq7));
			wq0 = vec_xor(wq0, wd0);
			wq1 = vec_xor(wq1, wd1);
			wq2 = vec_xor(wq2, wd2);
			wq3 = vec_xor(wq3, wd3);
			wq4 = vec_xor(wq4, wd4);
			wq5 = vec_xor(wq5, wd5);
			wq6 = vec_xor(wq6, wd6);
			wq7 = vec_xor(wq7, wd7);
		}
		*(unative_t *)&p[d+NSIZE*0] = wp0;
		*(unative_t *)&p[d+NSIZE*1] = wp1;
		*(unative_t *)&p[d+NSIZE*2] = wp2;
		*(unative_t *)&p[d+NSIZE*3] = wp3;
		*(unative_t *)&p[d+NSIZE*4] = wp4;
		*(unative_t *)&p[d+NSIZE*5] = wp5;
		*(unative_t *)&p[d+NSIZE*6] = wp6;
		*(unative_t *)&p[d+NSIZE*7] = wp7;
		*(unative_t *)&q[d+NSIZE*0] = wq0;
		*(unative_t *)&q[d+NSIZE*1] = wq1;
		*(unative_t *)&q[d+NSIZE*2] = wq2;
		*(unative_t *)&q[d+NSIZE*3] = wq3;
		*(unative_t *)&q[d+NSIZE*4] = wq4;
		*(unative_t *)&q[d+NSIZE*5] = wq5;
		*(unative_t *)&q[d+NSIZE*6] = wq6;
		*(unative_t *)&q[d+NSIZE*7] = wq7;
	}
}

static void raid6_vpermxor8_gen_syndrome(int disks, size_t bytes, void **ptrs)
{
	preempt_disable();
	enable_kernel_altivec();

	raid6_vpermxor8_gen_syndrome_real(disks, bytes, ptrs);

	disable_kernel_altivec();
	preempt_enable();
}

int raid6_have_altivec_vpermxor(void);
#if 8 == 1
int raid6_have_altivec_vpermxor(void)
{
	/* Check if arch has both altivec and the vpermxor instructions */
# ifdef __KERNEL__
	return (cpu_has_feature(CPU_FTR_ALTIVEC_COMP) &&
		cpu_has_feature(CPU_FTR_ARCH_207S));
# else
	return 1;
#endif

}
#endif

const struct raid6_calls raid6_vpermxor8 = {
	raid6_vpermxor8_gen_syndrome,
	NULL,
	raid6_have_altivec_vpermxor,
	"vpermxor8",
	0
};
#endif
